@@ -19,6 +19,8 @@
 #include "apreq_error.h"
 #include "apreq_util.h"
 
+#include "apr_lib.h" /* for apr_iscntrl() & co */
+
 #define PARSER_STATUS_CHECK(PREFIX)   do {         \
     if (ctx->status == PREFIX##_ERROR)             \
         return APREQ_ERROR_GENERAL;                \
@@ -39,7 +41,9 @@ struct hdr_ctx {
         HDR_GAP,
         HDR_VALUE,
         HDR_NEWLINE,
-        HDR_CONTINUE,
+        HDR_ENDLINE,
+        HDR_FOLDLINE,
+        HDR_LASTLINE,
         HDR_COMPLETE,
         HDR_ERROR
     }                   status;
@@ -48,27 +52,38 @@ struct hdr_ctx {
 /********************* header parsing utils ********************/
 
 
-static apr_status_t split_header_line(apreq_param_t **p,
-                                      apr_pool_t *pool,
-                                      apr_bucket_brigade *bb,
-                                      apr_size_t nlen,
-                                      apr_size_t glen,
-                                      apr_size_t vlen)
+static apr_bucket *split_header_line(apr_bucket *e, apr_size_t *off,
+                                     const char **data, apr_size_t *dlen)
+{
+    if (*off > 1) {
+        apr_bucket_split(e, *off - 1);
+        e = APR_BUCKET_NEXT(e);
+        *dlen -= *off - 1;
+        *data += *off - 1;
+        *off = 1;
+    }
+    return e;
+}
+
+static apr_status_t consume_header_line(apreq_param_t **p,
+                                        apr_pool_t *pool,
+                                        apr_bucket_brigade *bb,
+                                        apr_size_t nlen,
+                                        apr_size_t glen,
+                                        apr_size_t vlen)
 {
     apreq_param_t *param;
     apreq_value_t *v;
     apr_bucket *e, *f;
     apr_status_t s;
-    struct iovec vec[APREQ_DEFAULT_NELTS], *iov, *end;
+    struct iovec vec[APREQ_DEFAULT_NELTS], *iov;
     apr_array_header_t arr;
     char *dest;
     const char *data;
     apr_size_t dlen;
+    int i, eol = 0;
 
-    if (nlen == 0)
-        return APR_EBADARG;
-
-    param = apreq_param_make(pool, NULL, nlen, NULL, vlen - 1); /*drop (CR)LF */
+    param = apreq_param_make(pool, NULL, nlen, NULL, vlen);
     *(const apreq_value_t **)&v = &param->v;
 
     arr.pool     = pool;
@@ -80,69 +95,75 @@ static apr_status_t split_header_line(apreq_param_t **p,
     e = APR_BRIGADE_FIRST(bb);
 
     /* store name in a temporary iovec array */
-
-    while (nlen > 0) {
+    do {
         apr_size_t len;
-        end = apr_array_push(&arr);
-        s = apr_bucket_read(e, (const char **)&end->iov_base,
+
+        assert(e != APR_BRIGADE_SENTINEL(bb));
+        iov = (struct iovec *)apr_array_push(&arr);
+        s = apr_bucket_read(e, (const char **)&iov->iov_base,
                             &len, APR_BLOCK_READ);
         if (s != APR_SUCCESS)
             return s;
+        iov->iov_len = len;
 
         assert(nlen >= len);
-        end->iov_len = len;
         nlen -= len;
 
         e = APR_BUCKET_NEXT(e);
-    }
+    } while (nlen > 0);
 
     /* skip gap */
-
-    while (glen > 0) {
+    do {
+        assert(e != APR_BRIGADE_SENTINEL(bb));
         s = apr_bucket_read(e, &data, &dlen, APR_BLOCK_READ);
         if (s != APR_SUCCESS)
             return s;
 
         assert(glen >= dlen);
         glen -= dlen;
+
         e = APR_BUCKET_NEXT(e);
-    }
+    } while (glen > 0);
 
     /* copy value */
-    assert(vlen > 0);
     dest = v->data;
-    while (vlen > 0) {
+    do {
+        apr_size_t off;
 
+        assert(e != APR_BRIGADE_SENTINEL(bb));
         s = apr_bucket_read(e, &data, &dlen, APR_BLOCK_READ);
         if (s != APR_SUCCESS)
             return s;
 
-        memcpy(dest, data, dlen);
-        dest += dlen;
-        assert(vlen >= dlen);
-        vlen -= dlen;
+        for (off = 0; off < dlen; ++off) {
+            const char ch = data[off];
+            if (ch == '\r' || ch == '\n') {
+                /* Eat [CR]LF of continuation or end of line */
+                if (!vlen && ch == '\n')
+                    eol = 1; /* done */
+                continue;
+            }
+            assert(vlen > 0);
+            *dest = ch;
+            ++dest;
+            --vlen;
+        }
+
         e = APR_BUCKET_NEXT(e);
-    }
-
-    assert(dest[-1] == '\n');
-
-    if (dest[-2] == '\r')
-        --dest;
-
-    dest[-1] = 0;
-    v->dlen = (dest - v->data) - 1;
+    } while (!eol);
+    v->dlen = dest - v->data;
+    *dest++ = 0;
 
     /* write name */
     v->name = dest;
-    iov = (struct iovec *)arr.elts;
-
-    while (iov <= end) {
+    for (i = 0; i < arr.nelts; ++i) {
+        iov = &((struct iovec *)arr.elts)[i];
         memcpy(dest, iov->iov_base, iov->iov_len);
         dest += iov->iov_len;
         ++iov;
     }
+    v->nlen = dest - v->name;
     *dest = 0;
-    nlen = dest - v->name;
 
     while ((f = APR_BRIGADE_FIRST(bb)) != e)
         apr_bucket_delete(f);
@@ -153,12 +174,15 @@ static apr_status_t split_header_line(apreq_param_t **p,
 
 }
 
+#define IS_TOKEN_CHAR(c) (apr_isalnum(c) \
+                          || ((c) && strchr("!#$%&'*+-.^_`|~", (c))))
 
 APREQ_DECLARE_PARSER(apreq_parse_headers)
 {
     apr_pool_t *pool = parser->pool;
     apr_bucket *e;
     struct hdr_ctx *ctx;
+    char ch;
 
     if (parser->ctx == NULL) {
         ctx = apr_pcalloc(pool, sizeof *ctx);
@@ -173,9 +197,6 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
     e = APR_BRIGADE_LAST(ctx->bb);
     APR_BRIGADE_CONCAT(ctx->bb, bb);
 
- parse_hdr_brigade:
-
-
     /* parse the brigade for CRLF_CRLF-terminated header block,
      * each time starting from the front of the brigade.
      */
@@ -194,8 +215,8 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
             APR_BRIGADE_CONCAT(bb, ctx->bb);
             return APR_SUCCESS;
         }
-        s = apr_bucket_read(e, &data, &dlen, APR_BLOCK_READ);
 
+        s = apr_bucket_read(e, &data, &dlen, APR_BLOCK_READ);
         if ( s != APR_SUCCESS ) {
             ctx->status = HDR_ERROR;
             return s;
@@ -217,65 +238,57 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
         case HDR_NAME:
 
             while (off < dlen) {
-                switch (data[off++]) {
-
-                case '\n':
-                    if (off < dlen)
-                        apr_bucket_split(e, off);
-                    e = APR_BUCKET_NEXT(e);
-
-                    do {
-                        apr_bucket *f = APR_BRIGADE_FIRST(ctx->bb);
-                        apr_bucket_delete(f);
-                    } while (e != APR_BRIGADE_FIRST(ctx->bb));
-                    APR_BRIGADE_CONCAT(bb, ctx->bb);
-                    ctx->status = HDR_COMPLETE;
-                    return APR_SUCCESS;
-
+                ch = data[off++];
+                switch (ch) {
                 case ':':
-                    if (off > 1) {
-                        apr_bucket_split(e, off - 1);
-                        dlen -= off - 1;
-                        data += off - 1;
-                        off = 1;
-                        e = APR_BUCKET_NEXT(e);
+                    if (!ctx->nlen) {
+                        ctx->status = HDR_ERROR;
+                        return APREQ_ERROR_BADHEADER;
                     }
+                    e = split_header_line(e, &off, &data, &dlen);
                     ++ctx->glen;
                     ctx->status = HDR_GAP;
                     goto parse_hdr_bucket;
 
                 default:
+                    if (!IS_TOKEN_CHAR(ch)) {
+                        ctx->status = HDR_ERROR;
+                        return APREQ_ERROR_BADCHAR;
+                    }
                     ++ctx->nlen;
                 }
-
             }
-
             break;
 
 
         case HDR_GAP:
 
             while (off < dlen) {
-                switch (data[off++]) {
+                ch = data[off++];
+                switch (ch) {
                 case ' ':
                 case '\t':
                     ++ctx->glen;
                     break;
 
                 case '\n':
+                    e = split_header_line(e, &off, &data, &dlen);
                     ctx->status = HDR_NEWLINE;
                     goto parse_hdr_bucket;
 
+                case '\r':
+                    e = split_header_line(e, &off, &data, &dlen);
+                    ctx->status = HDR_ENDLINE;
+                    goto parse_hdr_bucket;
+
                 default:
-                    ctx->status = HDR_VALUE;
-                    if (off > 1) {
-                        apr_bucket_split(e, off - 1);
-                        dlen -= off - 1;
-                        data += off - 1;
-                        off = 1;
-                        e = APR_BUCKET_NEXT(e);
+                    if (apr_iscntrl(ch)) {
+                        ctx->status = HDR_ERROR;
+                        return APREQ_ERROR_BADCHAR;
                     }
+                    e = split_header_line(e, &off, &data, &dlen);
                     ++ctx->vlen;
+                    ctx->status = HDR_VALUE;
                     goto parse_hdr_bucket;
                 }
             }
@@ -285,60 +298,104 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
         case HDR_VALUE:
 
             while (off < dlen) {
-                ++ctx->vlen;
-                if (data[off++] == '\n') {
+                ch = data[off++];
+                switch (ch) {
+                case '\n':
                     ctx->status = HDR_NEWLINE;
                     goto parse_hdr_bucket;
+
+                case '\r':
+                    ctx->status = HDR_ENDLINE;
+                    goto parse_hdr_bucket;
+
+                default:
+                    if (apr_iscntrl(ch)) {
+                        ctx->status = HDR_ERROR;
+                        return APREQ_ERROR_BADCHAR;
+                    }
+                    ++ctx->vlen;
                 }
             }
             break;
 
 
+        case HDR_ENDLINE:
+        case HDR_LASTLINE:
+
+            if (off == dlen)
+                break;
+
+            if (data[off++] != '\n') {
+                ctx->status = HDR_ERROR;
+                return APREQ_ERROR_BADHEADER;
+            }
+
+            if (ctx->status == HDR_LASTLINE) {
+                ctx->status = HDR_COMPLETE;
+                goto parse_hdr_bucket;
+            }
+
+            /* fall thru */
+            ctx->status = HDR_NEWLINE;
+
         case HDR_NEWLINE:
 
             if (off == dlen)
                 break;
-            else {
-                switch (data[off]) {
 
-                case ' ':
-                case '\t':
-                    ctx->status = HDR_CONTINUE;
-                    ++off;
-                    ++ctx->vlen;
+            ch = data[off++];
+            switch (ch) {
+            case ' ':
+            case '\t':
+                ++ctx->vlen;
+                break;
+
+            default:
+                e = split_header_line(e, &off, &data, &dlen);
+
+                /* consume from splitted brigade now */
+                s = consume_header_line(&param, pool, ctx->bb,
+                                        ctx->nlen, ctx->glen, ctx->vlen);
+                if (parser->hook != NULL && s == APR_SUCCESS)
+                    s = apreq_hook_run(parser->hook, param, NULL);
+                if (s != APR_SUCCESS) {
+                    ctx->status = HDR_ERROR;
+                    return s;
+                }
+                apreq_value_table_add(&param->v, t);
+                ctx->nlen = 0;
+                ctx->vlen = 0;
+                ctx->glen = 0;
+
+                switch (ch) {
+                case '\n':
+                    ctx->status = HDR_COMPLETE;
+                    break;
+
+                case '\r':
+                    ctx->status = HDR_LASTLINE;
                     break;
 
                 default:
-                    /* can parse brigade now */
-                    if (off > 0)
-                        apr_bucket_split(e, off);
-                    s = split_header_line(&param, pool, ctx->bb, ctx->nlen, ctx->glen, ctx->vlen);
-                    if (parser->hook != NULL && s == APR_SUCCESS)
-                        s = apreq_hook_run(parser->hook, param, NULL);
-
-                    if (s != APR_SUCCESS) {
+                    if (!IS_TOKEN_CHAR(ch)) {
                         ctx->status = HDR_ERROR;
-                        return s;
+                        return APREQ_ERROR_BADCHAR;
                     }
-
-                    apreq_value_table_add(&param->v, t);
-                    e = APR_BRIGADE_SENTINEL(ctx->bb);
+                    ++ctx->nlen;
                     ctx->status = HDR_NAME;
-                    ctx->nlen = 0;
-                    ctx->vlen = 0;
-                    ctx->glen = 0;
-
-                    goto parse_hdr_brigade;
+                    break;
                 }
-
-                /* cases ' ', '\t' fall through to HDR_CONTINUE */
+                goto parse_hdr_bucket;
             }
 
+            /* fall thru */
+            ctx->status = HDR_FOLDLINE;
 
-        case HDR_CONTINUE:
+        case HDR_FOLDLINE:
 
             while (off < dlen) {
-                switch (data[off++]) {
+                ch = data[off++];
+                switch (ch) {
                 case ' ':
                 case '\t':
                     ++ctx->vlen;
@@ -348,7 +405,15 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
                     ctx->status = HDR_NEWLINE;
                     goto parse_hdr_bucket;
 
+                case '\r':
+                    ctx->status = HDR_ENDLINE;
+                    goto parse_hdr_bucket;
+
                 default:
+                    if (apr_iscntrl(ch)) {
+                        ctx->status = HDR_ERROR;
+                        return APREQ_ERROR_BADCHAR;
+                    }
                     ctx->status = HDR_VALUE;
                     ++ctx->vlen;
                     goto parse_hdr_bucket;
@@ -356,8 +421,24 @@ APREQ_DECLARE_PARSER(apreq_parse_headers)
             }
             break;
 
+
+        case HDR_COMPLETE:
+
+            if (off < dlen)
+                apr_bucket_split(e, off);
+
+            e = APR_BUCKET_NEXT(e);
+            do {
+                apr_bucket *f = APR_BRIGADE_FIRST(ctx->bb);
+                apr_bucket_delete(f);
+            } while (e != APR_BRIGADE_FIRST(ctx->bb));
+
+            APR_BRIGADE_CONCAT(bb, ctx->bb);
+            return APR_SUCCESS;
+
+
         default:
-            ; /* not reached */
+            assert(0); /* not reached */
         }
     }
     apreq_brigade_setaside(ctx->bb,pool);
