@@ -15,23 +15,22 @@
 **  limitations under the License.
 */
 
-#include "apr_file_io.h"
 #include "at.h"
-#include "apr_lib.h"
-#include "apr_strings.h"
-#include "apr_tables.h"
-#include "apr_env.h"
-#include <assert.h>
+#include <errno.h>
+#include <ctype.h>
 
+#define AT_SUCCESS 0
+#define AT_EGENERAL 14
+#define min(a, b) ((a < b) ? a : b)
 
-apr_status_t at_begin(at_t *t, int total)
+int at_begin(at_t *t, int total)
 {
     char buf[32];
-    apr_snprintf(buf, 32, "1..%d", total);
+    at_snprintf(buf, 32, "1..%d", total);
     return at_report(t, buf);
 }
 
-static apr_status_t test_cleanup(void *data)
+static int test_cleanup(void *data)
 {
     at_t *t = data;
     if (t->current < t->plan)
@@ -42,28 +41,28 @@ static apr_status_t test_cleanup(void *data)
 
 void at_end(at_t *t)
 {
-    apr_pool_cleanup_kill(t->pool, t, test_cleanup);
     test_cleanup(t);
+    free(t);
 }
 
-apr_status_t at_comment(at_t *t, const char *fmt, va_list vp)
+int at_comment(at_t *t, const char *fmt, va_list vp)
 {
-    apr_status_t s;
+    int s;
     char buf[256], *b = buf + 2;
     char *end;
     int rv;
-    rv = apr_vsnprintf(b, 250, fmt, vp);
+    rv = at_vsnprintf(b, 250, fmt, vp);
 
     if (rv <= 0)
-        return APR_EINVAL;
+        return EINVAL;
 
 
-    end = b + rv;
+    end = b + min(rv, 250);
 
     buf[0] = '#';
     buf[1] = ' ';
 
-    if (rv == 250) {
+    if (rv >= 250) {
         end[-1] = '.';
         *end++ = '.';
         *end++ = '.';
@@ -82,7 +81,7 @@ apr_status_t at_comment(at_t *t, const char *fmt, va_list vp)
         eol = strchr(b + 2, '\n');
         *eol = 0;
         s = at_report(t, b);
-        if (s != APR_SUCCESS || eol == end - 1)
+        if (s != AT_SUCCESS || eol == end - 1)
             break;
 
         b    = eol - 1;
@@ -118,38 +117,41 @@ void at_ok(at_t *t, int is_ok, const char *label, const char *file, int line)
         t->todo++;
         is_todo = 1;
     }
+    
+    if (AT_FLAG_TODO(t->flags))
+        is_todo = 1;
 
     if (AT_FLAG_CONCISE(t->flags))
         format[9] = '\0';
-    else if (is_ok && !AT_FLAG_TRACE(t->flags))
+    else if (is_ok && ! is_todo && !AT_FLAG_TRACE(t->flags))
         format[14] = '\0';
     else if (is_fatal && ! is_ok)
         comment = "fatal";
     else
         comment = is_todo ? "todo" : is_skip ? "skip" : "at";
 
-    rv = apr_snprintf(buf, 256, fmt, t->current + t->prior,
-                      label, comment,  file, line, t->current, t->name);
+    rv = at_snprintf(buf, 256, fmt, t->current + t->prior,
+                     label, comment,  file, line, t->current, t->name);
 
     if (rv <= 0)
         exit(-1);
 
-    end = buf + rv;
+    end = buf + min(rv, 250);
 
-    if (rv == 250) {
+    if (rv >= 250) {
         *end++ = '.';
         *end++ = '.';
         *end++ = '.';
         *end = '\0';
     }
 
-    if (memchr(buf, '\n', rv) != NULL || at_report(t, buf) != APR_SUCCESS)
+    if (memchr(buf, '\n', rv) != NULL || at_report(t, buf) != AT_SUCCESS)
         exit(-1);
 
     if (!is_ok && is_fatal) {
         while (t->current++ < t->plan) {
-            apr_snprintf(buf, 256, "not ok %d # skipped: aborting test %s",
-                     t->prior + t->current, t->name);
+            at_snprintf(buf, 256, "not ok %d # skipped: aborting test %s",
+                        t->prior + t->current, t->name);
             at_report(t, buf);
         }
         longjmp(*t->abort, 0);
@@ -158,68 +160,80 @@ void at_ok(at_t *t, int is_ok, const char *label, const char *file, int line)
 
 struct at_report_file {
     at_report_t module;
-    apr_file_t *file;
+    FILE *file;
 };
 
-
-static apr_status_t at_report_file_write(at_report_t *ctx, const char *msg)
+static int at_report_file_write(at_report_t *ctx, const char *msg)
 {
     struct at_report_file *r = (struct at_report_file *)ctx;
-    apr_file_t *f = r->file;
-    apr_size_t len = strlen(msg);
-    apr_status_t s;
+    FILE *f = r->file;
+    size_t len = strlen(msg);
+    size_t bytes_written;
+    int status;
 
-    s = apr_file_write_full(f, msg, len, &len);
-    if (s != APR_SUCCESS)
-        return s;
+    bytes_written = fwrite(msg, sizeof(char), len, f);
+    if (bytes_written != len)
+        return errno;
 
-    s = apr_file_putc('\n', f);
-    if (s != APR_SUCCESS)
-        return s;
+    status = putc('\n', f);
+    if (status == EOF)
+        return errno;
 
-    return apr_file_flush(f);
+    return fflush(f);
 }
 
-at_report_t *at_report_file_make(apr_pool_t *p, apr_file_t *f)
+at_report_t *at_report_file_make(FILE *f)
 {
-    struct at_report_file *r = apr_palloc(p, sizeof *r);
+    struct at_report_file *r = malloc(sizeof *r);
     r->module.func = at_report_file_write;
     r->file = f;
     return &r->module;
 }
 
-
+void at_report_file_cleanup(at_report_t *r)
+{
+    free(r);
+}
 
 struct at_report_local {
     at_report_t  module;
     at_t        *t;
     at_report_t *saved_report;
     const int   *saved_fatal;
+    const int   *saved_skip;
+    const int   *saved_todo;
     int          dummy_fatal;
-    const char  *file;
+    char        *file;
     int          line;
     int          passed;
-    apr_pool_t  *pool;
 };
 
-
-static apr_status_t report_local_cleanup(void *data)
+static int report_local_cleanup(void *data)
 {
     struct at_report_local *q = data;
     dAT = q->t;
     char label[32];
 
-    apr_snprintf(label, 32, "collected %d passing tests", q->passed);
+    at_snprintf(label, 32, "collected %d passing tests", q->passed);
 
     AT->report = q->saved_report;
-    AT->fatal = q->saved_fatal;
+    AT->fatal  = q->saved_fatal;
+    AT->skip   = q->saved_skip;
+    AT->todo   = q->saved_todo;
 
     at_ok(q->t, 1, label, q->file, q->line);
-    return APR_SUCCESS;
+
+    free(q->file);
+    free(q);
+
+    return AT_SUCCESS;
 }
 
+void at_report_delocalize(at_t *AT) {
+    report_local_cleanup(AT->report);
+}
 
-static apr_status_t at_report_local_write(at_report_t *ctx, const char *msg)
+static int at_report_local_write(at_report_t *ctx, const char *msg)
 {
     char buf[256];
     struct at_report_local *q = (struct at_report_local *)ctx;
@@ -227,12 +241,10 @@ static apr_status_t at_report_local_write(at_report_t *ctx, const char *msg)
 
     if (strncmp(msg, "not ok", 6) == 0) {
         q->saved_report->func(q->saved_report, msg);
-        AT->report = q->saved_report;
-        AT->fatal = q->saved_fatal;
-        apr_pool_cleanup_kill(q->pool, q, report_local_cleanup);
+        report_local_cleanup(q);
         while (AT->current++ < AT->plan) {
-            apr_snprintf(buf, 256, "not ok %d # skipped: aborting test %s",
-                     AT->prior + AT->current, AT->name);
+            at_snprintf(buf, 256, "not ok %d # skipped: aborting test %s",
+                        AT->prior + AT->current, AT->name);
             at_report(AT, buf);
         }
         longjmp(*AT->abort, 0);
@@ -241,76 +253,84 @@ static apr_status_t at_report_local_write(at_report_t *ctx, const char *msg)
         AT->current--;
         q->passed++;
     }
-    return APR_SUCCESS;
+    return AT_SUCCESS;
 }
 
-void at_report_local(at_t *AT, apr_pool_t *p, const char *file, int line)
+void at_report_local(at_t *AT, const char *file, int line)
 {
-    struct at_report_local *q = apr_palloc(p, sizeof *q);
+    struct at_report_local *q = malloc(sizeof *q);
+    size_t len;
+
     q->module.func = at_report_local_write;
     q->t = AT;
     q->saved_report = AT->report;
     q->saved_fatal = AT->fatal;
+    q->saved_skip = AT->skip;
+    q->saved_todo = AT->todo;
     q->dummy_fatal = 0;
-    q->file = apr_pstrdup(p, file);
     q->line = line;
     q->passed = 0;
-    q->pool = p;
 
-    AT->fatal = &q->dummy_fatal;
+    len = strlen(file) + 1;
+    q->file = (char*)malloc(len);
+    memcpy(q->file, file, len);
+
+    AT->fatal = AT->skip = AT->todo = &q->dummy_fatal;
     AT->report = &q->module;
 
     if (*q->saved_fatal == AT->current + 1)
         q->saved_fatal++;
 
-    apr_pool_cleanup_register(p, q, report_local_cleanup,
-                                    report_local_cleanup);
 }
 
-
-at_t *at_create(apr_pool_t *pool, unsigned char flags, at_report_t *report)
+at_t *at_create(unsigned char flags, at_report_t *report)
 {
-    at_t *t = apr_pcalloc(pool, sizeof *t);
+    at_t *t = calloc(sizeof *t, 1);
     t->flags = flags;
     t->report = report;
-    t->pool = pool;
-
-    apr_pool_cleanup_register(pool, t, test_cleanup, test_cleanup);
     return t;
 }
 
-
-#define AT_NELTS 4
-
-static int* at_list(apr_pool_t *pool, const char *spec, int *list)
+static int* at_list(const char *spec)
 {
-    apr_array_header_t arr;
-    int prev, current = 0;
-
-    arr.pool = pool;
-    arr.elt_size = sizeof *list;
-    arr.nelts = 0;
-    arr.nalloc = AT_NELTS;
-    arr.elts = (char *)list;
+    int prev, current = 0, count = 0, idx = 0, *elts;
+    const char *start = spec;
 
     do {
-        while (*spec && !apr_isdigit(*spec))
+        while (*spec && !isdigit((unsigned char)*spec))
             ++spec;
-
         prev = current;
-        current = (int)apr_strtoi64(spec, (char **)(void *)&spec, 10);
-        *(int *)apr_array_push(&arr) = current;
+        current = (int)strtol(spec, (char **)(void *)&spec, 10);
+        ++count;
 
-    } while (prev >= current);
+    } while (prev <= current);
 
-    return (int *)arr.elts;
+    elts = malloc(++count * sizeof *elts);
+    spec = start;
+    current = 0;
+
+    do {
+        while (*spec && !isdigit((unsigned char)*spec))
+            ++spec;
+        prev = current;
+        current = (int)strtol(spec, (char **)(void *)&spec, 10);
+        elts[idx++] = current;
+
+    } while (prev <= current);
+
+    elts[idx] = 0; /* sentinel */
+
+    return elts;
 }
 
+#define at_free_lists do { if (flist) free((void *)flist);       \
+                           if (slist) free((void *)slist);       \
+                           if (tlist) free((void *)tlist); } while (0)
 
-
-apr_status_t at_run(at_t *AT, const at_test_t *test)
+int at_run(at_t *AT, const at_test_t *test)
 {
-    int dummy = 0, fbuf[AT_NELTS], sbuf[AT_NELTS], tbuf[AT_NELTS];
+    int dummy = 0;
+    const int *flist = NULL, *slist = NULL, *tlist = NULL;
     jmp_buf j;
 
     AT->current = 0;
@@ -319,25 +339,56 @@ apr_status_t at_run(at_t *AT, const at_test_t *test)
     AT->plan    = test->plan;
 
     if (test->fatals)
-        AT->fatal = at_list(AT->pool, test->fatals, fbuf);
+        flist = AT->fatal = at_list(test->fatals);
     else
         AT->fatal = &dummy;
 
     if (test->skips)
-        AT->skip = at_list(AT->pool, test->skips, sbuf);
+        slist = AT->skip = at_list(test->skips);
     else
         AT->skip = &dummy;
 
     if (test->todos)
-        AT->todo = at_list(AT->pool, test->todos, tbuf);
+        tlist = AT->todo = at_list(test->todos);
     else
         AT->todo = &dummy;
 
     AT->abort = &j;
     if (setjmp(j) == 0) {
-        test->func(AT);
-        return APR_SUCCESS;
+        test->func(AT, test->ctx);
+        at_free_lists;
+        return AT_SUCCESS;
     }
+    at_free_lists;
     AT->abort = NULL;
-    return APR_EGENERAL;
+    return AT_EGENERAL;
 }
+
+int at_snprintf(char *buf, int size, const char *format, ...)
+{
+    va_list args;
+    int status;
+    va_start(args, format);
+    status = at_vsnprintf(buf, size, format, args);
+    va_end(args);
+    return status;
+}
+
+int at_vsnprintf(char *buf, int size, const char *format, va_list args)
+{
+#ifdef _MSC_VER
+    int status = _vsnprintf(buf, size, format, args);
+    /* Make Microsoft's _vsnprintf behave like C99 vsnprintf on overflow:
+     * NULL-terminate, and return the number of chars printed minus one. */
+    if (status < 0) {
+        if (errno != EINVAL) {
+            status = size - 1;
+            buf[status] = '\0';
+        }
+    }
+    return status;
+#else 
+    return vsnprintf(buf, size, format, args);
+#endif /* _MSC_VER */
+}
+
